
# patPlectrum.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>
Louis James Barman

"""


import MMA.notelen
import MMA.harmony

import gbl
from   MMA.common import *
from   MMA.pat import PC


class Plectrum(PC):
    """ Pattern class for a Raw MIDI track. """


    vtype = 'PLECTRUM'
    # We have vibrating strings (string in python refers to text in guitar string)
    _vibrating = []
    _tuning = []
    capoFretNo = 0  # The number that the capo is on (0 for open strings)

    def __init__(self, nm):
        PC.__init__(self, nm)
        self.setTuning("e-- a-- d- g- b- e")


    def addString(self, pitch):
        self._tuning.append(pitch)
        vibration = struct()
        vibration.note = 0       # 0 means the string not vibrating
        self._vibrating.append(vibration)


    def setTuning (self, stringsNames):
        #clear the old tuning
        del self._tuning[:]
        del self._vibrating[:]
        #  Tuning e-- a-- d- g- b- e
        self.addString(40)  #todo use the note names
        self.addString(45)
        self.addString(50)
        self.addString(55)
        self.addString(59)
        self.addString(64)


    def getPgroup(self, ev):
        """ Get group for rawmid pattern.

            Fields - start, length, note, volume

        """

        if not (len(ev) == 3 or len(ev) == len(self._tuning) + 2 ):
            error("There must be n groups of 3 or %d in a pattern definition, "
                  "not <%s>" % (len(self._tuning) + 2, ' '.join(ev)) )

        a = struct()

        a.offset   = self.setBarOffset(ev[0])
        a.strum = stoi(ev[1], "Strum in Plectrum definition not int")

        a.pluckVol = []
        if len(ev) == 3:
            vol = stoi(ev[2], "Note volume in Plectrum definition not int")
            for stringNo in range(len(self._tuning)):
                a.pluckVol.append(vol)
        else:
            for stringNo in range(len(self._tuning)):
                val = ev[2 + stringNo]
                if val=='-':
                    val = '-1'   # -1 means ignore this string
                a.pluckVol.append(stoi(val, "Note volume in Plectrum definition not int"))

        return a




    def restart(self):
        self.ssvoice = -1




    def endVibration(self, stringNo, offset):
        # first test if this srting has been played before
        if  self._vibrating[stringNo].note == 0:
            return

        vibration = self._vibrating[stringNo]
        gbl.mtrks[self.channel].addNoteOnToTrack( offset, vibration.note, 0) # when v=0 this is a note off
        #print "Note Off ", vibration.note
        self._vibrating[stringNo].note = 0



    # endAllVibrations (ie output all outstanding note off)
    def closeGroove(self, offset):
        for stringNo in self._vibrating:
            self.endVibration(stringNo, offset)



    def fretboardNote(self, stringNo, chordList, barreChordFretNo):
        openString = self._tuning[stringNo] + self.capoFretNo + barreChordFretNo
        lowestFretNo = 127
        for note in chordList:
            fretNo = note - openString % 12
            while fretNo < 0:
                fretNo += 12
            # find the lowest note
            if fretNo < lowestFretNo:
                lowestFretNo = fretNo

        return openString + lowestFretNo



    def trackBar(self, pattern, ctable):
        """ Do the bass bar.

        Called from self.bar()

        """

        sc = self.seq
        unify = self.unify[sc]


        for p in pattern:
            ct = self.getChordInPos(p.offset, ctable)

            chordList=ct.chord.noteList

            # Todo
            #if ct.bassZ:
            #    continue

            if len(self._tuning) != len(p.pluckVol):
                error("Wrong number of strings")
                continue

            plectrumNoteOnList = [] # for debugging only
            previousNote = 0


            # Find how many strings have been plucked this time
            pluckStringCount = 0;
            for stringNo, vol in enumerate(p.pluckVol):
                if vol == -1:
                    continue
                pluckStringCount += 1


            pluckStringIndex = 0;
            for stringNo, vol in enumerate(p.pluckVol):
                if vol == -1:
                    continue

                # find the centre of the strum
                strumOffset = p.offset + p.strum*(pluckStringIndex - pluckStringCount/2.0);
                pluckStringIndex += 1

                self.endVibration(stringNo, strumOffset)
                if vol >= 1:
                    note = self.fretboardNote(stringNo, chordList, 0)

                    if note == previousNote:
                        warn("duplicate note in the chord Igrnoring")
                        continue
                    previousNote = note

                    outputVolume = self.adjustVolume(vol, p.offset)

                    gbl.mtrks[self.channel].addNoteOnToTrack(strumOffset, note, outputVolume)

                    self._vibrating[stringNo].note = note

                    plectrumNoteOnList.append(note)  # for debugging only


            if gbl.debug:
                print "chordList ", chordList,  " NoteOn ", plectrumNoteOnList, " p.offset ", p.offset









